/**
 * An example on how to create a synchronized game that uses a global tick count 
 */

import 'package:webrtc_utils/client.dart';
import 'package:webrtc_utils/game.dart';
import 'dart:html';
import 'dart:math';
import 'dart:async';

final String webSocketUrl = 'ws://${window.location.hostname}:28080';

const TICKS_PER_SECONDS = 10;

/**
 * Demo Shape that will exchanged / changed synchronously 
 */

class Circle {
  final Point middle;
  
  Circle(this.middle);
  
  int _opacity = 100;
  
  int _radius = 0;
  
  bool tick() {
    _opacity -= 4;
    _radius += 1;
    return _opacity <= 0;
  }
  
  void render(CanvasRenderingContext2D ctx) {
    ctx.save();
    ctx.fillStyle = 'red';
    ctx.globalAlpha = _opacity / 100.0;
    ctx.beginPath();
    ctx.arc(middle.x, middle.y, _radius, 0, 2 * PI);
    ctx.fill();
    ctx.restore();
  }
}

/**
 * Shared logic
 */

abstract class DemoPlayer {
  SynchronizedGameRoom get room;
  
  List<Circle> circles = [];
  
  /**
   * Listener on the game room that waits for synchronized messages
   */
  
  void _setupListener() {
    room.onSynchronizedMessage.listen((Map message) {
      Point p = new Point(message['click']['x'], message['click']['y']);
      _click(p);
    });
  }
  
  void _click(Point p) {
    circles.add(new Circle(p));
  }
  
  void tick(int tick) {
    List<Circle> finishedCircles = [];
    circles.forEach((Circle c) {
      if(c.tick()) {
        finishedCircles.add(c);
      }
    });
    circles.removeWhere((Circle c) => finishedCircles.contains(c));
  }
  
  void render(CanvasRenderingContext2D ctx) {
    circles.forEach((Circle c) => c.render(ctx));
  }
}

/**
 * An example of a local player
 */

class MyLocalPlayer extends SynchronizedLocalPlayer with DemoPlayer {
  CanvasElement _canvas;
  
  MyLocalPlayer(SynchronizedGameRoom gameRoom, int id) : super(gameRoom, id) {
    _setupListener();
    _canvas = querySelector('#canvas');
    _canvas.onClick.listen((MouseEvent ev) {
      gameRoom.synchronizeMessage({'click': {'x': ev.offset.x, 'y': ev.offset.y}, 'time': (window.performance.now() + gameRoom.maxPing * 2)});
    });
  }
}

/**
 * An example of a remote player
 */

class MyRemotePlayer extends SynchronizedRemotePlayer with DemoPlayer {
  MyRemotePlayer(SynchronizedGameRoom room, Peer peer) : super(room, peer) {
    _setupListener();
  }
}

/**
 * Protocol provider that handles message serialization
 */

class MyProtocolProvider extends DefaultProtocolProvider {
  DataChannelProtocol provide(Peer peer, RtcDataChannel channel) {
    if(channel.protocol == 'game') {
      return new JsonProtocol(channel);
    }
    
    return super.provide(peer, channel);
  }
}

/**
 * A synchronized game example. Overrides only the createGameRoom method so
 * you can create a game specific room
 */

class TickPlayer {
  
}

class TickGame<L extends TickPlayer, R extends TickPlayer> {
      
}

class SynchronizedGame extends SynchronizedP2PGame<MyLocalPlayer,MyRemotePlayer> with TickGame<MyLocalPlayer, MyRemotePlayer>
  /* with AlivePlayerGame<MyLocalPlayer, MyRemotePlayer>*/ {
  
  CanvasElement canvas;
  
  CanvasRenderingContext2D ctx;
  
  int fpsCounter = 0;
  
  SynchronizedGame(String webSocketUrl, Map rtcConfiguration)
    : super(webSocketUrl, rtcConfiguration) {
    setProtocolProvider(new MyProtocolProvider());
    
    canvas = querySelector('#canvas');
    ctx = canvas.getContext('2d');
    startAnimation();
    new Timer.periodic(new Duration(seconds: 1), (Timer t) {
      querySelector('#fps').text = '$fpsCounter';
      fpsCounter = 0;
    });
  }
  
  // TODO(rh): Do not initialize to 0! Wait for onSynchronized event and take global time for initialization
  // double _lastTickTime = 0.0;
  int lastTick = 0;
  
  void _tick(double localTime) {
    fpsCounter++;
    // Make tick in each room to update the room-global time
    gameRooms.forEach((SynchronizedGameRoom room) {
      room.tick(localTime);
    });
    
    // Rendering...
    // TODO(rh): Decouple these two steps!
    ctx.clearRect(0,0, canvas.width, canvas.height);
    gameRooms.forEach((SynchronizedGameRoom room) {
      double globalTime = room.timeDifferenceToMaster + localTime;
      int currentTick = globalTime ~/ TICKS_PER_SECONDS;
      querySelector('#globaltime').text = '${globalTime}';
      room.players.forEach((Player player) {
        for(int t=lastTick+1;t<=currentTick;t++) {
          player.tick(t);
        }
        if(player is DemoPlayer) {
          (player as DemoPlayer).render(ctx);
        }
      });
      
      lastTick = currentTick;
    });
    
    startAnimation();
  }
  
  void startAnimation() {
    window.requestAnimationFrame(_tick);
  }
  
  @override
  MyLocalPlayer createLocalPlayer(GameRoom room, int localId) {
    return new MyLocalPlayer(room, localId);
  }
  
  @override
  MyRemotePlayer createRemotePlayer(GameRoom room, Peer peer) {
    return new MyRemotePlayer(room, peer);
  }
}

/**
 * Main method that will create the [SynchronizedGame] and
 * join a room once connected to the signaling server
 */

void main() {
  final SynchronizedP2PGame game = new SynchronizedGame(webSocketUrl, rtcConfiguration);
  
  game.onConnect.listen((_) {
    print('Connected');
    game.join('room1');
  });
}